==> dynamiq-0.2.0.tar.gz: dynamiq-0.2.0/dynamiq/memory/backend/in_memory.py <==
import math
from collections import Counter

from dynamiq.memory.backend.base import MemoryBackend
from dynamiq.prompts import Message


class InMemoryError(Exception):
    """Base exception class for InMemory backend errors."""

    pass


class BM25DocumentRanker:
    """BM25 implementation for scoring documents."""

    def __init__(self, documents: list[str], k1: float = 1.5, b: float = 0.75):
        """Initialize with a list of documents and parameters for BM25."""
        self.documents = documents
        self.k1 = k1
        self.b = b
        self.avg_dl = self._calculate_avg_dl()

    def _calculate_avg_dl(self):
        """Calculates the average document length (number of terms per document)."""
        total_length = sum(len(doc.lower().split()) for doc in self.documents)
        return total_length / len(self.documents) if self.documents else 0

    def _idf(self, term: str, N: int, df: int) -> float:
        """Calculates the IDF (inverse document frequency) of a term."""
        return math.log((N - df + 0.5) / (df + 0.5) + 1)

    def score(self, query_terms: list[str], document: str) -> float:
        """Calculates the BM25 score for a document."""
        doc_terms = document.lower().split()
        doc_len = len(doc_terms)
        doc_term_freqs = Counter(doc_terms)
        N = len(self.documents)
        score = 0.0

        for term in query_terms:
            term_freq = doc_term_freqs.get(term, 0)
            if term_freq == 0:
                continue
            df = sum(1 for doc in self.documents if term in doc.lower().split())
            idf = self._idf(term, N, df)
            numerator = term_freq * (self.k1 + 1)
            denominator = term_freq + self.k1 * (1 - self.b + self.b * (doc_len / self.avg_dl))
            score += idf * (numerator / denominator)

        return score


class InMemory(MemoryBackend):
    """In-memory implementation of the memory storage backend."""

    name = "InMemory"

    def __init__(self):
        """Initializes the in-memory storage."""
        self.messages: list[Message] = []

    def add(self, message: Message):
        """Adds a message to the in-memory list."""
        try:
            self.messages.append(message)
        except Exception as e:
            raise InMemoryError(f"Error adding message to InMemory backend: {e}") from e

    def get_all(self) -> list[Message]:
        """Retrieves all messages from the in-memory list."""
        return sorted(self.messages, key=lambda msg: msg.metadata.get("timestamp", 0))  # Sort by timestamp

    def _apply_filters(self, messages: list[Message], filters: dict) -> list[Message]:
        """Applies metadata filters to the list of messages."""
        if not filters:
            return messages
        filtered_messages = messages
        for key, value in filters.items():
            if isinstance(value, list):
                filtered_messages = [msg for msg in filtered_messages if any(v == msg.metadata.get(key) for v in value)]
            else:
                filtered_messages = [msg for msg in filtered_messages if value == msg.metadata.get(key)]
        return filtered_messages

    def search(self, query: str = None, limit: int = None, filters: dict = None) -> list[Message]:
        """Searches for messages using BM25 scoring, with optional filters."""
        if not query and not filters:
            return self.get_all()[:limit]

        filtered_messages = self._apply_filters(self.messages, filters)
        if not query:
            return filtered_messages[:limit]
        query_terms = query.lower().split()
        document_texts = [msg.content for msg in filtered_messages]
        bm25 = BM25DocumentRanker(document_texts)
        scored_messages = [(msg, bm25.score(query_terms, msg.content)) for msg in filtered_messages]
        scored_messages = [(msg, score) for msg, score in scored_messages if score > 0]
        scored_messages.sort(key=lambda x: x[1], reverse=True)
        return [msg for msg, _ in scored_messages][:limit]

    def is_empty(self) -> bool:
        """Checks if the in-memory list is empty."""
        return len(self.messages) == 0

    def clear(self):
        """Clears the in-memory list."""
        self.messages = []

==> dynamiq-0.3.2.tar.gz: dynamiq-0.3.2/dynamiq/memory/backend/in_memory.py <==
import math
from collections import Counter

from dynamiq.memory.backend.base import MemoryBackend
from dynamiq.prompts import Message


class InMemoryError(Exception):
    """Base exception class for InMemory backend errors."""

    pass


class BM25DocumentRanker:
    """BM25 implementation for scoring documents."""

    def __init__(self, documents: list[str], k1: float = 1.5, b: float = 0.75):
        """Initialize with a list of documents and parameters for BM25."""
        self.documents = documents
        self.k1 = k1
        self.b = b
        self.avg_dl = self._calculate_avg_dl()

    def _calculate_avg_dl(self):
        """Calculates the average document length (number of terms per document)."""
        total_length = sum(len(doc.lower().split()) for doc in self.documents)
        return total_length / len(self.documents) if self.documents else 0

    def _idf(self, term: str, N: int, df: int) -> float:
        """Calculates the IDF (inverse document frequency) of a term."""
        return math.log((N - df + 0.5) / (df + 0.5) + 1)

    def score(self, query_terms: list[str], document: str) -> float:
        """Calculates the BM25 score for a document."""
        doc_terms = document.lower().split()
        doc_len = len(doc_terms)
        doc_term_freqs = Counter(doc_terms)
        N = len(self.documents)
        score = 0.0

        for term in query_terms:
            term_freq = doc_term_freqs.get(term, 0)
            if term_freq == 0:
                continue
            df = sum(1 for doc in self.documents if term in doc.lower().split())
            idf = self._idf(term, N, df)
            numerator = term_freq * (self.k1 + 1)
            denominator = term_freq + self.k1 * (1 - self.b + self.b * (doc_len / self.avg_dl))
            score += idf * (numerator / denominator)

        return score


class InMemory(MemoryBackend):
    """In-memory implementation of the memory storage backend."""

    name = "InMemory"

    def __init__(self):
        """Initializes the in-memory storage."""
        self.messages: list[Message] = []

    def add(self, message: Message):
        """Adds a message to the in-memory list."""
        try:
            self.messages.append(message)
        except Exception as e:
            raise InMemoryError(f"Error adding message to InMemory backend: {e}") from e

    def get_all(self) -> list[Message]:
        """Retrieves all messages from the in-memory list."""
        return sorted(self.messages, key=lambda msg: msg.metadata.get("timestamp", 0))  # Sort by timestamp

    def _apply_filters(self, messages: list[Message], filters: dict) -> list[Message]:
        """Applies metadata filters to the list of messages."""
        if not filters:
            return messages
        filtered_messages = messages
        for key, value in filters.items():
            if isinstance(value, list):
                filtered_messages = [msg for msg in filtered_messages if any(v == msg.metadata.get(key) for v in value)]
            else:
                filtered_messages = [msg for msg in filtered_messages if value == msg.metadata.get(key)]
        return filtered_messages

    def search(self, query: str = None, limit: int = 10, filters: dict = None) -> list[Message]:
        """Searches for messages using BM25 scoring, with optional filters."""
        if not query and not filters:
            return self.get_all()[:limit]

        filtered_messages = self._apply_filters(self.messages, filters)
        if not query:
            return filtered_messages[:limit]
        query_terms = query.lower().split()
        document_texts = [msg.content for msg in filtered_messages]
        bm25 = BM25DocumentRanker(document_texts)
        scored_messages = [(msg, bm25.score(query_terms, msg.content)) for msg in filtered_messages]
        scored_messages = [(msg, score) for msg, score in scored_messages if score > 0]
        scored_messages.sort(key=lambda x: x[1], reverse=True)
        return [msg for msg, _ in scored_messages][:limit]

    def is_empty(self) -> bool:
        """Checks if the in-memory list is empty."""
        return len(self.messages) == 0

    def clear(self):
        """Clears the in-memory list."""
        self.messages = []
